import numpy as np
import pandas as pd
import scanpy as sc
import scvelo as scv
import warnings
import matplotlib.pyplot as plt

warnings.filterwarnings("ignore")


def get_child_dict(adata, group_key, path_dict, threshold=0.1, threshold_trans=1.0, adjust=False):
    '''
    :param adata: anndata to use
    :param group_key: cell cluster key to use
    :param path_dict: path dictionary from recover direction
    :param threshold: threshold for weight of edges
    :param threshold_trans: threshold for product of weights of edges
    :return: child node dictionary for each cluster node, level dictionary for each cluster node
    :return: if conduct adjustment for t_2 in DTI
    '''

    # rerun paga and get full tree
    sc.tl.diffmap(adata)
    sc.tl.paga(adata, groups=group_key)

    weight = adata.uns["paga"]["connectivities"].A
    cluster_name = list(adata.obs[group_key].values.categories)
    start_node = path_dict["start_node"]

    # create level dictionary for data
    thres_list = []
    threshold = threshold * np.max(weight)
    level_dict = {}
    level = 0
    parent_nodes = []
    child_nodes = [start_node]
    child_nodes_ = []
    while len(child_nodes) > 0:
        for node in child_nodes:
            level_dict[node] = level
            i = cluster_name.index(node)
            for i_ in np.where(weight[i, :] > threshold)[0]:
                node_ = cluster_name[i_]
                weight_trans = max([weight[i, j] * weight[j, i_] for j in range(len(cluster_name))])
                thres_list.append(weight[i, i_] / weight_trans)
                if node_ not in parent_nodes + child_nodes and threshold_trans * weight_trans <= weight[i, i_]:
                    child_nodes_.append(node_)

        parent_nodes = parent_nodes + child_nodes
        child_nodes = child_nodes_.copy()
        child_nodes_ = []
        level += 1
        thres_list.sort()

    # get child nodes for every node
    child_dict = {}
    for node in cluster_name:
        child_list = []
        for i in np.where(weight[cluster_name.index(node), :] > threshold)[0]:
            node_ = cluster_name[i]
            try:
                if level_dict[node] == level_dict[node_] - 1:
                    child_list.append(node_)
            except KeyError:
                pass
        child_dict[node] = child_list

    for node in cluster_name:
        if node != start_node and not np.any([node in child_dict[node_] for node_ in cluster_name]):
            child_dict[cluster_name[np.argmax(weight[cluster_name.index(node), :])]] += [node]

    for node in cluster_name:
        parent_nodes = [node_ for node_ in cluster_name if node in child_dict[node_]]
        if len(parent_nodes) > 1:
            i = cluster_name.index(node)
            parent_index = [cluster_name.index(node_) for node_ in parent_nodes]
            weight_node_ = [weight[i, j] for j in parent_index]
            node_ = cluster_name[parent_index[np.argmax(weight_node_)]]
            for node__ in cluster_name:
                if node__ != node_ and node in child_dict[node__]:
                    child_dict[node__].remove(node)

    if not adjust:
        adata.uns["child_dict"] = child_dict.copy()
        adata.uns["level_dict"] = level_dict.copy()
        adata.uns["threshold_list"] = thres_list.copy()
        print("'child_dict' added to adata.uns")
        print("'level_dict' added to adata.uns")
        print("'threshold_list' added to adata.uns")
    return child_dict, level_dict, thres_list


# create directed nearest neighborhood
def get_d_nn(adata, group_key, child_dict, start_node, emb_key=None, root_select="connectivities"):

    start_cell_idx = np.flatnonzero(adata.obs[group_key] == start_node)
    child_cell_idx = []
    for child_node in child_dict[start_node]:
        child_cell_idx += list(np.flatnonzero(adata.obs[group_key] == child_node))

    if root_select == "connectivities":
        root_cell_idx = np.flatnonzero(adata.obs[group_key] == start_node)[0]
    elif root_select == "coordinates" and emb_key is not None:
        coordinates = adata.obsm[emb_key]
        start_xy = coordinates[start_cell_idx, :]
        child_xy = coordinates[child_cell_idx, :]
        dist_xy = np.sum(np.square(start_xy), axis=1, keepdims=True) + np.sum(np.square(child_xy), axis=1, keepdims=True).T \
                  - 2 * np.dot(start_xy, child_xy.T)
        root_cell_idx = start_cell_idx[np.argmin(np.sum(dist_xy, axis=1))]
    elif root_select == "terminal_score":
        root_score = adata.obs["root_cells"]
        start_root_score = root_score[start_cell_idx]
        root_cell_idx = start_cell_idx[np.argmax(start_root_score)]
    elif root_select == "coordinates" and emb_key is None:
        raise KeyError("Embedding key not provided")
    else:
        raise KeyError("No such method for root cell selection")

    adata.uns['iroot'] = root_cell_idx
    sc.tl.dpt(adata)
    dpt_pseudotime = adata.obs["dpt_pseudotime"].values
    cluster_label = np.array(adata.obs[group_key].values)
    cluster_name = list(adata.obs[group_key].values.categories)

    knn = (adata.obsp["distances"].A != 0).astype(float)
    child_c, same_c = np.zeros_like(knn), np.zeros_like(knn)
    for cluster in cluster_name:
        child_node = child_dict[cluster]
        if_child = np.array([j in child_node for j in cluster_label])
        if_same = np.array([j in [cluster] for j in cluster_label])
        child_c[cluster_label == cluster, :] = if_child
        same_c[cluster_label == cluster, :] = if_same

    d_nn = knn * (child_c + same_c) * (dpt_pseudotime[:, None] <= dpt_pseudotime)
    adata.obsp["d_nn"] = d_nn.copy()
    print("'d_nn' added to adata.obsp")

    return d_nn, knn, child_c, same_c


def directed_graph(adata, child_dict, group_key, start_node, emb_key="X_umap", data_name=None, show=False, ax=None):
    # visualize the directed graph
    embedding = adata.obsm[emb_key]
    adata.obs["{}0".format(emb_key)] = embedding[:, 0]
    adata.obs["{}1".format(emb_key)] = embedding[:, 1]
    info = pd.concat([adata.obs.groupby(group_key)["{}0".format(emb_key)].mean(),
                      adata.obs.groupby(group_key)["{}1".format(emb_key)].mean(),
                      pd.Series(adata.obs[group_key].value_counts(), name="ncells")], axis=1)

    groups_sizes = adata.obs[group_key].value_counts()
    median_group_size = np.median(groups_sizes)
    groups_sizes = pd.Series(120 * np.power(groups_sizes / median_group_size, 0.5), name="sizes")
    info = pd.concat([info, groups_sizes], axis=1)

    try:
        colors = pd.Series(adata.uns[group_key + "_colors"], index=info.index, name="colors")
        info = pd.concat([info, colors], axis=1)
    except KeyError:
        pass

    weight = adata.uns["paga"]["connectivities"].A
    weight = weight + weight.T
    cluster_name = list(adata.obs[group_key].values.categories)

    if ax is None:
        fig, ax = plt.subplots(1, 1)

    start_index = list(info.index).index(start_node)
    remain = info.drop(start_node, axis=0)
    ax.scatter(remain["{}0".format(emb_key)], remain["{}1".format(emb_key)], s=remain['sizes'], c="tab:blue",
               zorder=1)
    ax.scatter(info["{}0".format(emb_key)][start_index], info["{}1".format(emb_key)][start_index],
               s=info['sizes'][start_index], c="orange", zorder=0)

    for i in range(weight.shape[0]):
        for j in range(weight.shape[1]):
            if cluster_name[j] in child_dict[cluster_name[i]]:
                x, y = info["{}0".format(emb_key)][i], info["{}1".format(emb_key)][i]
                u, v = info["{}0".format(emb_key)][j], info["{}1".format(emb_key)][j]
                ax.quiver(x, y, u-x, v-y, angles='xy', scale_units='xy', scale=1, zorder=1, color="gray")

    for count, cluster_name in enumerate(info.index):
        ax.text(info["{}0".format(emb_key)][count], info["{}1".format(emb_key)][count], cluster_name,
                ha="left", zorder=2)

    ax.set_xticks([])
    ax.set_yticks([])

    if data_name is not None:
        plt.title(data_name)
    if show:
        plt.show()

    return ax










import os
import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import scvelo as scv
from .path.process import process_path
from .direction.correct import correct_path
from .velocity.DTI import get_child_dict, get_d_nn, directed_graph
from .velocity.model import get_velocity
from .utils.metrics import inner_cluster_coh, cross_boundary_correctness, cross_boundary_scvelo_probs,\
    cross_boundary_correctness2, inner_cluster_coh2, velo_coh


def tivelo(adata, group_key, emb_key, res=0.6, data_name="data", save_folder="results", njobs=-1,
           start_mode="stochastic", rev_stat="mean", tree_gene=None, t1=0.1, t2=1, show_fig=True, filter_genes=True,
           constrain=True, loss_fun="mse", only_s=True, alpha_1=1, alpha_2=0.1, batch_size=1024, n_epochs=100,
           adjust_DTI=False, show_DTI=False, velocity_key="velocity", cluster_edges=None, measure_performance=True):

    # create path
    result_path = save_folder + "/{}/".format(data_name)
    figs_path = save_folder + "/{}/figs/".format(data_name)
    if not os.path.exists(result_path):
        os.makedirs(result_path)
    if not os.path.exists(figs_path):
        os.makedirs(figs_path)

    # load existing adata
    try:
        adata_ = sc.read_h5ad(result_path + "tivelo.h5ad")

        # load dictionary
        child_dict_ = adata_.uns["child_dict"]
        path_dict = adata_.uns["path_dict"]
        start_node = adata_.uns["path_dict"]["start_node"]
        thres_list = adata_.uns["threshold_list"]

        # add PAGA result to original adata
        if "paga" not in adata_.uns.keys():
            sc.tl.diffmap(adata_)
            sc.tl.paga(adata_, groups=group_key)

    except FileNotFoundError:
        # set group_key, emb_key and neighbourhood
        if group_key is None or group_key not in adata.obs.keys():
            sc.tl.leiden(adata, resolution=res)
            group_key = "leiden"
        if "neighbors" not in adata.uns.keys():
            sc.pp.neighbors(adata, n_neighbors=30, n_pcs=30)
            sorted_indices = np.argsort(adata.obsp["distances"].A + np.identity(adata.n_obs), axis=1)
            sorted_indices = np.fliplr(sorted_indices)
            indices = sorted_indices[:, 0: 30]
            adata.uns["neighbors"]["indices"] = indices
        if "indices" not in adata.uns["neighbors"].keys():
            sorted_indices = np.argsort(adata.obsp["distances"].A + np.identity(adata.n_obs), axis=1)
            sorted_indices = np.fliplr(sorted_indices)
            indices = sorted_indices[:, 0: 30]
            adata.uns["neighbors"]["indices"] = indices
        if emb_key is None or emb_key not in adata.obsm.keys():
            scv.tl.umap(adata)
            emb_key = "X_umap"

        # step 1
        path_dict, ax = process_path(adata, group_key, emb_key, njobs=njobs, start_mode=start_mode)

        ax.axis('off')
        ax.legend(loc="best", ncol=1, prop={'size': 10}, frameon=False)
        plt.savefig(figs_path + "path.pdf")
        if show_fig:
            plt.show()
        plt.close()

        # step 2
        path_dict, ax = correct_path(path_dict, adata, group_key, rev_stat=rev_stat, tree_gene=tree_gene,
                                     root_select="connectivities")

        if ax is not None:
            plt.savefig(figs_path + "tree.pdf")
            if show_fig:
                plt.show()
            plt.close()

        # step 3
        # get DNN
        child_dict_, level_dict_, thres_list = get_child_dict(adata, group_key, path_dict, threshold=t1,
                                                              threshold_trans=t2, adjust=False)
        start_node = list(level_dict_.keys())[0]
        d_nn_, knn_, child_c_, same_c_ = get_d_nn(adata, group_key, child_dict_, start_node, emb_key=emb_key,
                                                  root_select="connectivities")

        # velocity inference
        adata_, v_u, v_s = get_velocity(adata, d_nn_, knn_, same_c=same_c_, n_epochs=n_epochs, loss_fun=loss_fun,
                                        only_s=only_s, constrain=constrain, alpha_1=alpha_1, alpha_2=alpha_2,
                                        batch_size=batch_size, filter_genes=filter_genes)

        scv.tl.velocity_graph(adata_, vkey=velocity_key, n_jobs=njobs)
        scv.tl.velocity_embedding(adata_)
        adata_.write_h5ad(result_path + "tivelo.h5ad")

    # velocity stream plot
    ax = scv.pl.velocity_embedding_stream(adata_, vkey=velocity_key, color=group_key, title='', cutoff_perc=0,
                                          show=False)
    plt.savefig(figs_path + "tivelo.png")
    if show_fig:
        plt.show()
    plt.close()

    # DTI
    if show_DTI:
        ax = directed_graph(adata_, child_dict_, group_key, start_node, emb_key=emb_key, data_name=data_name)
        ax.axis('off')
        plt.savefig(figs_path + "DTI.pdf")
        if show_fig:
            plt.show()
        plt.close()

    # adjust DTI
    if adjust_DTI:
        for t_ in thres_list:
            child_dict_new, _, _ = get_child_dict(adata_, group_key, path_dict, threshold=t1, threshold_trans=t_,
                                                  adjust=False)
            if child_dict_new != child_dict_:
                child_dict_ = child_dict_new
                ax = directed_graph(adata_, child_dict_, group_key, start_node, emb_key=emb_key, data_name=data_name)
                ax.set_title(data_name + ", " + r"$t_2$={}".format(np.round(t_, 2)))
                plt.savefig(figs_path + "/DTI_t2={}.png".format(np.round(t_, 2)))
                if show_fig:
                    plt.show()

    # measure performance
    if measure_performance:
        if cluster_edges is not None:
            _, cbdir = cross_boundary_correctness(adata_, cluster_key=group_key, velocity_key=velocity_key,
                                                  cluster_edges=cluster_edges, x_emb=emb_key)
            _, cbdir2 = cross_boundary_correctness2(adata_, cluster_key=group_key, velocity_key=velocity_key,
                                                    cluster_edges=cluster_edges)
            _, trans_probs = cross_boundary_scvelo_probs(adata_, cluster_key=group_key, cluster_edges=cluster_edges,
                                                         trans_g_key="{}_graph".format(velocity_key))
            _, icvcoh = inner_cluster_coh(adata_, cluster_key=group_key, velocity_key=velocity_key)
            _, icvcoh2 = inner_cluster_coh2(adata_, cluster_key=group_key, velocity_key=velocity_key, x_emb=emb_key)
            velocoh = velo_coh(adata_, velocity_key=velocity_key, trans_g_key="{}_graph".format(velocity_key))

            print("TIVelo:\n", "CBDir:", "%.4f" % cbdir, "ICVCoh:", "%.4f" % icvcoh, "\n",
                  "CBDir2:", "%.4f" % cbdir2, "ICVCoh2:", "%.4f" % icvcoh2, "\n",
                  "TransProbs:", "%.4f" % trans_probs, "VeloCoh:", "%.4f" % velocoh)

        else:
            _, icvcoh = inner_cluster_coh(adata_, cluster_key=group_key, velocity_key=velocity_key)
            _, icvcoh2 = inner_cluster_coh2(adata_, cluster_key=group_key, velocity_key=velocity_key, x_emb=emb_key)
            velocoh = velo_coh(adata_, velocity_key=velocity_key, trans_g_key="{}_graph".format(velocity_key))

            print("TIVelo:\n", "ICVCoh:", "%.4f" % icvcoh, "\n", "ICVCoh2:", "%.4f" % icvcoh2, "\n",
                  "VeloCoh:", "%.4f" % velocoh)

    return adata_






import numpy as np
import scanpy as sc
import matplotlib.pyplot as plt
from .prune import graph_prune


def get_child_index(select_weight, parent_index):
    # given a parent node index i, return child(i)
    child_index = []
    for index in parent_index:
        index_ = list(np.where(select_weight[index, :] > 0)[0])
        child_index += index_
    return list(set(child_index))


def get_child_path(select_weight, parent_index, root_index):
    # given root node index i and its one child node index i' (parent_index), return successor(i)
    child_index_ = get_child_index(select_weight, parent_index)
    child_index_ = list(set(child_index_) - set(root_index))
    child_path = child_index_
    while len(child_index_) > 0:
        child_index = child_index_
        root_index += child_index
        child_index_ = get_child_index(select_weight, child_index)
        child_index_ = list(set(child_index_) - set(root_index))
        child_path += child_index_
    return child_path


def path_selection(select_info, select_weight, start_index, root_list=None):
    '''
    :param select_info: dataframe information for selected clusters and edges
    :param select_weight: weight matrix for selected clusters and edges
    :param start_index: start node index
    :param root_list: list of root cluster excluded from selection
    :return: path of nodes beginning from start node excluding nodes in root list, with most number of cells
    '''
    select_path = [start_index]
    root_index = [start_index] + root_list if root_list is not None else [start_index]
    split_index = list(np.where(select_weight[start_index, :] > 0)[0])

    while len(split_index) > 0:
        ncells_list = []
        split_index = list(np.where(select_weight[start_index, :] > 0)[0])
        split_index = list(set(split_index) - set(root_index))
        if len(split_index) == 0:
            break

        for node_index in split_index:
            child_path = get_child_path(select_weight, [node_index], [start_index, node_index])
            ncells = select_info["ncells"][[node_index] + child_path].sum()
            ncells_list.append(ncells)

        select_index = split_index[np.argmax(ncells_list)]
        select_path += [select_index]
        start_index = select_index
        root_index += [select_index]

    return select_path


def get_path(select_info, select_weight, start_index):
    '''
    :param select_info: dataframe information for selected clusters and edges
    :param select_weight: weight matrix for selected clusters and edges
    :param start_index: start node index
    :return: dictionary of selected main path including start node and following nodes
    '''
    path_dict = {}
    root_list = []

    select_path = path_selection(select_info, select_weight, start_index, root_list=None)
    main_path = list(select_info.index[select_path].values)
    path_dict["start_node"] = select_info.index[start_index]
    path_dict["main_path"] = main_path

    root_list += select_path
    flag = 1
    for index in select_path:
        branch_index = list(set(get_child_index(select_weight, [index])) - set(root_list))
        if len(branch_index) > 0:
            for index_ in branch_index:
                select_path = get_child_path(select_weight, [index_], root_list + [index_])
                select_path = [index] + [index_] + select_path
                path_dict["branch_{}".format(flag)] = list(select_info.index[select_path].values)
                flag += 1

    return path_dict


def visualize_path(select_info, select_weight, path_dict, emb_key="X_umap", show=True, ax=None):
    flag1, flag2 = 0, 0

    # special adjustment for organoid:
    if "Enteroendocrine progenitor" in list(select_info.index):
        stem_idx = list(select_info.index).index("Stem cells")
        select_info["{}1".format(emb_key)][stem_idx] = select_info["{}1".format(emb_key)][stem_idx] + 0.5

        ta_idx = list(select_info.index).index("TA cells")
        select_info["{}1".format(emb_key)][ta_idx] = select_info["{}1".format(emb_key)][ta_idx] + 0.5

    if ax is None:
        fig, ax = plt.subplots(1, 1)
    for i in range(select_weight.shape[0]):
        for j in range(select_weight.shape[1]):
            if select_weight[i, j] > 0:
                x = select_info["{}0".format(emb_key)][i], select_info["{}0".format(emb_key)][j]
                y = select_info["{}1".format(emb_key)][i], select_info["{}1".format(emb_key)][j]
                if select_info.index[i] in path_dict["main_path"] and select_info.index[j] in path_dict["main_path"]:
                    if flag1 == 0:
                        ax.plot(x, y, lw=5 * select_weight[i, j], c="red", zorder=0, label="Main path")
                        flag1 += 1
                    else:
                        ax.plot(x, y, lw=5 * select_weight[i, j], c="red", zorder=0)
                else:
                    if flag2 == 0:
                        ax.plot(x, y, lw=5 * select_weight[i, j], c="green", zorder=0, label="Branch")
                        flag2 += 1
                    else:
                        ax.plot(x, y, lw=5 * select_weight[i, j], c="green", zorder=0)

    start_index = list(select_info.index).index(path_dict["start_node"])
    remain = select_info.drop(path_dict["start_node"], axis=0)
    ax.scatter(remain["{}0".format(emb_key)], remain["{}1".format(emb_key)], s=remain['sizes'], c="tab:blue", zorder=1,
               )
    ax.scatter(select_info["{}0".format(emb_key)][start_index], select_info["{}1".format(emb_key)][start_index],
               s=select_info['sizes'][start_index], c="orange", zorder=1, label="Origin node")

    for count, cluster_name in enumerate(select_info.index):
        ax.text(select_info["{}0".format(emb_key)][count], select_info["{}1".format(emb_key)][count], cluster_name,
                ha="center", zorder=2, fontsize=12)
    if show:
        plt.show()
    else:
        return ax


if __name__ == '__main__':
    # test
    data_path = r"D:\cuhk\project\velocity\dataset\scRNA-seq\BoneMarrow\human_cd34_bone_marrow_processed.h5ad"
    group_key = "clusters"
    emb_key = "X_tsne"

    adata = sc.read_h5ad(data_path)
    select_info, select_weight = graph_prune(adata, group_key, emb_key)

    path_dict = get_path(select_info, select_weight, start_index=5)
    print(path_dict)
    visualize_path(select_info, select_weight, path_dict, emb_key, show=True)







import numpy as np
import scanpy as sc
import matplotlib.pyplot as plt
from .linear_tree import create_tree, create_forecast
from ..utils.velocity_genes import compute_velocity_genes


def get_slope(tree, data):
    if not isinstance(tree, dict):
        return tree[1]
    if data[tree['spInd']] > tree['spVal']:
        if isinstance(tree, dict):
            return get_slope(tree['left'], data)
        else:
            return tree[1]
    else:
        if isinstance(tree, dict):
            return get_slope(tree['right'], data)
        else:
            return tree[1]


def rle(inarray):
    # for a sequence, return the corresponding run-length, position and element
    ia = np.asarray(inarray)
    n = len(ia)
    if n == 0:
        return None, None, None
    else:
        y = np.array(ia[1:] != ia[:-1])
        i = np.append(np.where(y), n - 1)
        z = np.diff(np.append(-1, i))
        p = np.cumsum(np.append(0, z))[:-1]
        return z, p, ia[i]


def orient_score(u, s, rate=1e-10, dur=100, kernel_size=100, step=1):
    n_cells = len(u)

    # preprocess
    u = np.convolve(u, np.ones(kernel_size) / kernel_size, mode="same")
    s = np.convolve(s, np.ones(kernel_size) / kernel_size, mode="same")
    if np.sum(u) > 0:
        u = u / np.sum(u)
    if np.sum(s) > 0:
        s = s / np.sum(s)

    x = np.arange(n_cells)
    u_arr = np.concatenate((x[:, None], u[:, None]), axis=1)
    s_arr = np.concatenate((x[:, None], s[:, None]), axis=1)

    # create tree
    u_tree = create_tree(u_arr, rate, dur, step)
    s_tree = create_tree(s_arr, rate, dur, step)

    # calculate slope
    u_slope, s_slope = np.zeros(n_cells), np.zeros(n_cells)
    for i in range(n_cells):
        u_slope[i] = get_slope(u_tree, u_arr[i, :])
        s_slope[i] = get_slope(s_tree, s_arr[i, :])

    # select sections
    u_rise, s_rise = u_slope > 0, s_slope > 0
    u_fall, s_fall = u_slope < 0, s_slope < 0
    rise_sec = u_rise & s_rise
    fall_sec = u_fall & s_fall

    # calculate run length, normalized reads, weight of location for rising section
    zr, pr, er = rle(rise_sec)
    pr_ = np.append(pr, n_cells)
    rise_l, rise_w, u_norm, s_norm = np.zeros(n_cells), np.zeros(n_cells), np.zeros(n_cells), np.zeros(n_cells)
    for i in range(len(pr)):
        rise_l[pr_[i]: pr_[i+1]] = zr[i]
        rise_w[pr_[i]: pr_[i+1]] = pr[i]
        u_norm[pr_[i]: pr_[i+1]] = u[pr_[i]: pr_[i+1]] - u[pr_[i]]
        s_norm[pr_[i]: pr_[i + 1]] = s[pr_[i]: pr_[i + 1]] - s[pr_[i]]

    rise_diff = u - s

    # calculate run length, normalized reads, weight of location for falling section
    zf, pf, ef = rle(fall_sec)
    pf_ = np.append(pf, n_cells)
    fall_l, fall_w, u_norm, s_norm = np.zeros(n_cells), np.zeros(n_cells), np.zeros(n_cells), np.zeros(n_cells)
    for i in range(len(pf)):
        fall_l[pf_[i]: pf_[i+1]] = zf[i]
        fall_w[pf_[i]: pf_[i+1]] = pf[i]
        u_norm[pf_[i]: pf_[i+1]] = u[pf_[i]: pf_[i+1]] - u[pf_[i]]
        s_norm[pf_[i]: pf_[i + 1]] = s[pf_[i]: pf_[i + 1]] - s[pf_[i]]

    fall_diff = u - s

    # calculate orientation scores
    pos_score = np.sum(rise_sec * rise_l * rise_diff)
    neg_score = np.sum(fall_sec * fall_l * fall_diff)
    score = pos_score - neg_score

    # calculate section score
    rise_p_ = [pr[i] for i in np.arange(len(pr)) if er[i] == 1]
    rise_l_ = [zr[i] for i in np.arange(len(zr)) if er[i] == 1]
    pos_score_ = [np.sum(((u_rise & s_rise) * rise_l * rise_diff)[pr_[i]: pr_[i+1]]) for i in
                  np.arange(len(pr)) if er[i] == 1]
    fall_p_ = [pf[i] for i in np.arange(len(pf)) if ef[i] == 1]
    fall_l_ = [zf[i] for i in np.arange(len(zf)) if ef[i] == 1]
    neg_score_ = [np.sum(((u_fall & s_fall) * fall_l * fall_diff)[pf_[i]: pf_[i + 1]]) for i in
                   np.arange(len(pf)) if ef[i] == 1]

    sec_list = [rise_sec, fall_sec, rise_p_, fall_p_, rise_l_, fall_l_, pos_score_, neg_score_]

    return score, u_arr, s_arr, u_tree, s_tree, sec_list


def visualize(u_arr, s_arr, u_tree, s_tree, sec_list, fig_size=(12, 9)):
    u_hat = create_forecast(u_tree, u_arr[:, 0])
    s_hat = create_forecast(s_tree, s_arr[:, 0])
    rise_sec, fall_sec, rise_p_, fall_p_, rise_l_, fall_l_, pos_score_, neg_score_ = sec_list

    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=fig_size, sharex=True)

    ax1.plot(u_arr[:, 0], u_arr[:, 1], c="blue", label="Mu")
    ax1.plot(s_arr[:, 0], s_arr[:, 1], c="red", label="Ms")

    # search for the beginning points
    n_cells = len(u_arr)
    u_slope, s_slope = np.zeros(n_cells), np.zeros(n_cells)
    for i in range(n_cells):
        u_slope[i] = get_slope(u_tree, u_arr[i, :])
        s_slope[i] = get_slope(s_tree, s_arr[i, :])
    zu, pu, eu = rle(u_slope)
    zs, ps, es = rle(s_slope)
    pu_ = np.append(pu, n_cells)
    ps_ = np.append(ps, n_cells)
    # plot
    for i in np.arange(len(pu)):
        if i == 0:
            ax2.plot(u_arr[pu_[i]:pu_[i + 1], 0], np.maximum(u_hat[pu_[i]:pu_[i + 1]], 0), c="blue", label="Fitted Mu")
        else:
            ax2.plot(u_arr[pu_[i]:pu_[i + 1], 0], np.maximum(u_hat[pu_[i]:pu_[i + 1]], 0), c="blue")
        ax2.plot(u_arr[pu[i]-1:pu[i]+1, 0], np.maximum(u_hat[pu[i]-1:pu[i]+1], 0), c="blue", ls="dotted")
    for i in np.arange(len(ps)):
        if i == 0:
            ax2.plot(s_arr[ps_[i]:ps_[i + 1], 0], np.maximum(s_hat[ps_[i]:ps_[i + 1]], 0), c="red", label="Fitted Ms")
        else:
            ax2.plot(s_arr[ps_[i]:ps_[i + 1], 0], np.maximum(s_hat[ps_[i]:ps_[i + 1]], 0), c="red")
        ax2.plot(s_arr[ps[i]-1:ps[i]+1, 0], np.maximum(s_hat[ps[i] - 1:ps[i]+1], 0), c="red", ls="dotted")

    max_val = max(np.max(u_arr[:, 1]), np.max(s_arr[:, 1]))
    ax1.fill_between(np.arange(len(rise_sec)), 0, max_val, where=rise_sec, color="orange", alpha=0.1, label="Rising")
    ax1.fill_between(np.arange(len(fall_sec)), 0, max_val, where=fall_sec, color="green", alpha=0.1, label="Falling")
    # ax1.text(0, max_val, "Total Score: {:.2f}".format(score))
    for i in range(len(rise_p_)):
        ax2.text(rise_p_[i] + rise_l_[i] / 2, 0, "{:.2f}".format(pos_score_[i]), ha="center", va="bottom", fontsize=12)
    for i in range(len(fall_p_)):
        ax2.text(fall_p_[i] + fall_l_[i] / 2, 0, "{:.2f}".format(-neg_score_[i]), ha="center", va="bottom", fontsize=12)

    max_val = max(np.max(u_hat), np.max(s_hat))
    ax2.fill_between(np.arange(len(rise_sec)), 0, max_val, where=rise_sec, color="orange", alpha=0.1)
    ax2.fill_between(np.arange(len(fall_sec)), 0, max_val, where=fall_sec, color="green", alpha=0.1)

    return fig, (ax1, ax2)














import numpy as np
from sklearn.metrics.pairwise import cosine_similarity


def acc(y_true, y_pred):
    """
    Calculate clustering accuracy. Require scikit-learn installed
    # Arguments
        y: true labels, numpy.array with shape `(n_samples,)`
        y_pred: predicted labels, numpy.array with shape `(n_samples,)`
    # Return
        accuracy, in [0,1]
    """
    y_true = y_true.astype(np.int64)
    assert y_pred.size == y_true.size
    D = max(y_pred.max(), y_true.max()) + 1
    w = np.zeros((D, D), dtype=np.int64)
    for i in range(y_pred.size):
        w[y_pred[i], y_true[i]] += 1
    from scipy.optimize import linear_sum_assignment
    ind = linear_sum_assignment(w.max() - w)
    ind = np.asarray(ind)
    ind = np.transpose(ind)
    return sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size


def keep_type(adata, nodes, target, cluster_key):

    return nodes[adata.obs[cluster_key][nodes].values == target]


def cross_boundary_correctness(adata, cluster_key, velocity_key, cluster_edges, return_raw=False, x_emb="X_umap"):

    scores = {}
    all_scores = {}

    if x_emb == "X_umap":
        v_emb = adata.obsm['{}_umap'.format(velocity_key)]
    else:
        v_emb = adata.obsm[[key for key in adata.obsm if key.startswith(velocity_key)][0]]

    x_emb = adata.obsm[x_emb]

    for u, v in cluster_edges:
        sel = adata.obs[cluster_key] == u
        nbs = adata.uns['neighbors']['indices'][sel]

        boundary_nodes = map(lambda nodes: keep_type(adata, nodes, v, cluster_key), nbs)
        x_points = x_emb[sel]
        x_velocities = v_emb[sel]

        type_score = []
        for x_pos, x_vel, nodes in zip(x_points, x_velocities, boundary_nodes):
            if len(nodes) == 0:
                continue

            position_dif = x_emb[nodes] - x_pos
            dir_scores = cosine_similarity(position_dif, x_vel.reshape(1, -1)).flatten()
            type_score.append(np.mean(dir_scores))

        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score

    if return_raw:
        return all_scores

    return scores, np.mean([sc for sc in scores.values() if not np.isnan(sc)])


def cross_boundary_correctness2(adata, cluster_key, velocity_key, cluster_edges, return_raw=False):

    scores = {}
    all_scores = {}

    velocities = adata.layers[velocity_key]
    x = adata.layers["Ms"]

    for u, v in cluster_edges:
        sel = adata.obs[cluster_key] == u
        nbs = adata.uns['neighbors']['indices'][sel]

        boundary_nodes = map(lambda nodes: keep_type(adata, nodes, v, cluster_key), nbs)
        x_points = x[sel]
        x_velocities = velocities[sel]

        type_score = []
        for x_pos, x_vel, nodes in zip(x_points, x_velocities, boundary_nodes):
            if len(nodes) == 0:
                continue

            position_dif = x[nodes] - x_pos
            dir_scores = cosine_similarity(position_dif, x_vel.reshape(1, -1)).flatten()
            type_score.append(np.mean(dir_scores))

        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score

    if return_raw:
        return all_scores

    return scores, np.mean([sc for sc in scores.values() if not np.isnan(sc)])


def inner_cluster_coh(adata, cluster_key, velocity_key, return_raw=False):

    clusters = np.unique(adata.obs[cluster_key])
    scores = {}
    all_scores = {}
    for cat in clusters:
        sel = adata.obs[cluster_key] == cat
        nbs = adata.uns['neighbors']['indices'][sel]
        same_cat_nodes = map(lambda nodes: keep_type(adata, nodes, cat, cluster_key), nbs)
        velocities = adata.layers[velocity_key]
        cat_vels = velocities[sel]
        cat_score = [cosine_similarity(cat_vels[[ith]], velocities[nodes]).mean()
                     for ith, nodes in enumerate(same_cat_nodes)
                     if len(nodes) > 0]
        all_scores[cat] = cat_score
        scores[cat] = np.mean(cat_score)

    if return_raw:
        return all_scores

    return scores, np.mean([sc for sc in scores.values() if not np.isnan(sc)])


def inner_cluster_coh2(adata, cluster_key, velocity_key, return_raw=False, x_emb="X_umap"):
    scores = {}
    all_scores = {}

    if x_emb == "X_umap":
        v_emb = adata.obsm['{}_umap'.format(velocity_key)]
    else:
        v_emb = adata.obsm[[key for key in adata.obsm if key.startswith(velocity_key)][0]]

    clusters = np.unique(adata.obs[cluster_key])
    for cat in clusters:
        sel = adata.obs[cluster_key] == cat
        nbs = adata.uns['neighbors']['indices'][sel]
        same_cat_nodes = map(lambda nodes: keep_type(adata, nodes, cat, cluster_key), nbs)
        cat_vels = v_emb[sel]
        cat_score = [cosine_similarity(cat_vels[[ith]], v_emb[nodes]).mean()
                     for ith, nodes in enumerate(same_cat_nodes)
                     if len(nodes) > 0]
        all_scores[cat] = cat_score
        scores[cat] = np.mean(cat_score)

    if return_raw:
        return all_scores

    return scores, np.mean([sc for sc in scores.values() if not np.isnan(sc)])


def cross_boundary_scvelo_probs(adata, cluster_key, cluster_edges, trans_g_key, return_raw=False):
    scores = {}
    all_scores = {}

    for u, v in cluster_edges:
        sel = adata.obs[cluster_key] == u
        nbs = adata.uns['neighbors']['indices'][sel]
        boundary_nodes = map(lambda nodes: keep_type(adata, nodes, v, cluster_key), nbs)
        type_score = [trans_probs.toarray()[:, nodes].mean()
                      for trans_probs, nodes in zip(adata.uns[trans_g_key][sel], boundary_nodes)
                      if len(nodes) > 0]
        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score
    if return_raw:
        return all_scores
    return scores, np.mean([sc for sc in scores.values() if not np.isnan(sc)])


def cross_boundary_coh(adata, cluster_key, velocity_key, cluster_edges, return_raw=False):
    scores = {}
    all_scores = {}
    for u, v in cluster_edges:
        sel = adata.obs[cluster_key] == u
        nbs = adata.uns['neighbors']['indices'][sel]
        boundary_nodes = map(lambda nodes: keep_type(adata, nodes, v, cluster_key), nbs)

        velocities = adata.layers[velocity_key]
        v_us = velocities[sel]
        type_score = [cosine_similarity(v_us[[ith]], velocities[nodes]).mean()
                      for ith, nodes in enumerate(boundary_nodes)
                      if len(nodes) > 0]
        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score

    if return_raw:
        return all_scores

    return scores, np.mean([sc for sc in scores.values() if not np.isnan(sc)])


def velo_coh(adata, velocity_key, trans_g_key):
    s = adata.layers["Ms"]
    T = adata.uns[trans_g_key].toarray()
    T = np.exp(T) / np.sum(np.exp(T), axis=1, keepdims=True)
    Ts = np.dot(T, s)
    delta = Ts - s
    velocities = adata.layers[velocity_key]

    velo_coh = np.sum(delta * velocities, axis=1) / (np.linalg.norm(delta, axis=1) * np.linalg.norm(velocities, axis=1) + 1e-8)
    return np.mean(velo_coh)





import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import norm
from sklearn.neighbors import NearestNeighbors


def gaussian_kernel(X, mu=0, sigma=1):
    return np.exp(-(X - mu) ** 2 / (2 * sigma ** 2)) / np.sqrt(2 * np.pi * sigma ** 2)


def sampling_neighbors(gene_unsplice_splice, step=(30, 30), percentile=25):
    grs = []
    for dim_i in range(gene_unsplice_splice.shape[1]):
        m, M = np.min(gene_unsplice_splice[:, dim_i]), np.max(gene_unsplice_splice[:, dim_i])
        m = m - 0.025 * np.abs(M - m)
        M = M + 0.025 * np.abs(M - m)
        gr = np.linspace(m, M, step[dim_i])
        grs.append(gr)
    meshes_tuple = np.meshgrid(*grs)
    gridpoints_coordinates = np.vstack([i.flat for i in meshes_tuple]).T
    gridpoints_coordinates = gridpoints_coordinates + norm.rvs(loc=0, scale=0.15, size=gridpoints_coordinates.shape)

    np.random.seed(42)  # set random seed

    nn = NearestNeighbors()

    neighbors_1 = min((gene_unsplice_splice[:, 0:2].shape[0] - 1), 20)
    nn.fit(gene_unsplice_splice[:, 0:2])
    dist, ixs = nn.kneighbors(gridpoints_coordinates, neighbors_1)

    ix_choice = ixs[:, 0].flat[:]
    ix_choice = np.unique(ix_choice)

    nn = NearestNeighbors()

    neighbors_2 = min((gene_unsplice_splice[:, 0:2].shape[0] - 1), 20)
    nn.fit(gene_unsplice_splice[:, 0:2])
    dist, ixs = nn.kneighbors(gene_unsplice_splice[ix_choice, 0:2], neighbors_2)

    density_extimate = gaussian_kernel(dist, mu=0, sigma=0.5).sum(1)
    bool_density = density_extimate > np.percentile(density_extimate, percentile)
    ix_choice = ix_choice[bool_density]

    return (ix_choice)


def scatter_gene(adata, velocity_key="velocity", gene_name=None, ax=None, color=None, custom_xlim=None,
                 custom_ylim=None, vmin=None, vmax=None, alpha=1, s=10, velocity=True, arrow_grid=(15, 15),
                 key_cluster=None, length_scale=5):

    if color in adata.obs.keys() and color + "_colors" in adata.uns.keys():
        group_list = list(adata.obs[color].cat.categories)
        color_list = adata.uns[color + "_colors"]
        color_dict = dict(zip(group_list, color_list))
        if key_cluster not in list(adata.obs[color].cat.categories):
            c = [color_dict[category] for category in adata.obs[color]]
        else:
            c = ["#FF0000" if category == key_cluster else "#95D9EF" for category in adata.obs[color]]

    elif color is None:
        c = "#95D9EF"

    assert gene_name, '\nError! gene is required!\n'

    gene_idx = list(adata.var_names).index(gene_name)
    Mu, Ms = adata.layers['Mu'][:, gene_idx], adata.layers['Ms'][:, gene_idx]
    ax.scatter(Ms, Mu, c=c, s=s, alpha=alpha, vmin=vmin, vmax=vmax, edgecolor="none")

    if custom_xlim is not None:
        ax.set_xlim(custom_xlim[0], custom_xlim[1])
    if custom_ylim is not None:
        ax.set_ylim(custom_ylim[0], custom_ylim[1])

    if velocity:
        u_s = np.concatenate((Mu[:, None], Ms[:, None]), axis=1)
        u_max, s_max = np.max(Mu), np.max(Ms)

        velocity = adata.layers[velocity_key][:, gene_idx]
        velocity_u = adata.layers[velocity_key + "_u"][:, gene_idx]
        # velocity_u, velocity = np.clip(velocity_u, -u_max/20, u_max/20), np.clip(velocity, -s_max/10, s_max/10)

        # preprocess velocity vector using sigmoid function
        velocity_u = u_max / length_scale * (2 / (1 + np.exp(-velocity_u)) - 1)
        velocity = s_max / length_scale * (2 / (1 + np.exp(-velocity)) - 1)

        sampling_idx = sampling_neighbors(u_s, step=arrow_grid, percentile=5)  # Sampling
        u_s_downsample = u_s[sampling_idx, :]
        velocity_downsample = velocity[sampling_idx]
        velocity_u_downsample = velocity_u[sampling_idx]

        plt.scatter(u_s_downsample[:, 1], u_s_downsample[:, 0], color="none", s=s, edgecolor="k")
        plt.quiver(u_s_downsample[:, 1], u_s_downsample[:, 0],
                   velocity_downsample, velocity_u_downsample,
                   angles='xy', scale_units='xy', scale=1)

    return ax